import json
import os
import uuid
from typing import Text
from unittest.mock import Mock

import pytest
from _pytest.monkeypatch import MonkeyPatch
from rasax.community.constants import RASA_PRODUCTION_ENVIRONMENT
from ruamel.yaml import YAMLError
from ruamel.yaml.compat import ordereddict

import rasax.community.utils
from rasa.nlu.config import InvalidConfigError
from rasax.community import config, utils
from rasax.community.services.settings_service import (
    SettingsService,
    default_stack_config,
    default_environments_config_local,
    ProjectException,
)


def default_environments_config():
    stack_config = ordereddict(
        [
            (
                RASA_PRODUCTION_ENVIRONMENT,
                ordereddict(
                    [
                        ("url", "http://rasa-production:5005"),
                        ("token", config.rasa_token),
                    ]
                ),
            ),
            (
                "development",
                ordereddict(
                    [
                        ("url", "http://rasa-development:5005"),
                        ("token", config.rasa_token),
                    ]
                ),
            ),
            (
                "worker",
                ordereddict(
                    [("url", "http://rasa-worker:5005"), ("token", config.rasa_token)]
                ),
            ),
        ]
    )

    return {"environments": ordereddict([("rasa", stack_config)])}


def test_project_init(settings_service: SettingsService):
    project_name = str(uuid.uuid4())
    settings_service.init_project(config.team_name, project_name)

    actual = settings_service.get(config.team_name, project_name)

    assert actual["config"] == default_stack_config()


def test_project_init_already_exists(settings_service: SettingsService):
    # create new project and check it exists
    project = "new project"
    team = "random"
    settings_service.init_project(team, project)
    assert settings_service.get(team, project)

    # creating the same project again fails
    with pytest.raises(ProjectException):
        settings_service.init_project(team, project)


def test_save_config(
    settings_service: SettingsService, monkeypatch: MonkeyPatch, new_project: Text
):
    test_config = default_stack_config()
    test_config["extra"] = "extra value"

    dump_config = Mock()
    monkeypatch.setattr(settings_service, "_dump_config", dump_config)
    monkeypatch.setattr(utils, "should_dump", lambda: True)
    settings_service.save_config(config.team_name, new_project, test_config)

    dump_config.assert_called_once()
    actual = settings_service.get_config(config.team_name, new_project)

    assert actual == test_config


def test_save_config_but_dont_dump(
    settings_service: SettingsService, monkeypatch: MonkeyPatch, new_project: Text
):
    test_config = default_stack_config()
    test_config["extra"] = "extra value"

    dump_config = Mock()
    monkeypatch.setattr(settings_service, "_dump_config", dump_config)
    settings_service.save_config(
        config.team_name, new_project, test_config, should_dump=False
    )

    dump_config.assert_not_called()
    actual = settings_service.get_config(config.team_name, new_project)

    assert actual == test_config


def test_get_config_for_invalid_project(settings_service: SettingsService):
    test_config = default_stack_config()
    test_config["extra"] = "extra value"

    assert not settings_service.get_config(
        "one person is not a team", "not existing project"
    )


def test_get_stack_services(settings_service: SettingsService, new_project: Text):
    default_env = default_environments_config()["environments"]["rasa"]

    settings_service.save_environments_config(
        new_project, default_environments_config().get("environments")
    )

    actual = settings_service.stack_services(new_project)

    assert len(actual) == len(default_env)

    for k, v in default_env.items():
        assert actual[k].stack_endpoint.token == v["token"]
        assert actual[k].stack_endpoint.url == v["url"]


@pytest.mark.parametrize(
    "environment_config",
    [default_environments_config(), default_environments_config_local(1234)],
)
def test_inspect_valid_environment_config(environment_config):
    default_env = environment_config["environments"]

    assert SettingsService.inspect_environments_config(default_env) is None


def test_invalid_environment_config_without_rasa_key():
    default_env = default_environments_config()["environments"]
    default_env["other_key"] = default_env.pop("rasa")

    with pytest.raises(ValueError):
        SettingsService.inspect_environments_config(default_env)


@pytest.mark.parametrize("environment", [RASA_PRODUCTION_ENVIRONMENT, "worker"])
def test_invalid_environment_config_without_minimal_instances(environment):
    default_env = default_environments_config()["environments"]
    del default_env["rasa"][environment]

    with pytest.raises(ValueError):
        SettingsService.inspect_environments_config(default_env)


def test_extra_environments_with_enterprise():
    default_env = default_environments_config()["environments"]
    default_env["rasa"]["another"] = default_env["rasa"][RASA_PRODUCTION_ENVIRONMENT]

    assert SettingsService.inspect_environments_config(default_env) is None


def test_extra_environments_with_rasa_ce():
    default_env = default_environments_config()["environments"]
    default_env["rasa"]["another"] = default_env["rasa"][RASA_PRODUCTION_ENVIRONMENT]

    # path to pretend we don't have enterprise installed
    rasax.community.utils.is_enterprise_installed = lambda: False

    with pytest.raises(ValueError):
        SettingsService.inspect_environments_config(default_env)

    # revert patch
    rasax.community.utils.is_enterprise_installed = lambda: True


def test_save_environment_config(settings_service: SettingsService):
    default_env = default_environments_config()["environments"]
    test_project = "default"

    settings_service.save_environments_config(test_project, default_env)

    assert settings_service.get_environments_config(test_project)
    assert len(settings_service.stack_services(test_project)) == len(
        default_environments_config()["environments"]["rasa"]
    )


def test_delete_environment_config(settings_service: SettingsService):
    default_env = default_environments_config()["environments"]
    test_project = "default"

    settings_service.save_environments_config(test_project, default_env)

    settings_service.delete_environment_config(
        RASA_PRODUCTION_ENVIRONMENT, test_project
    )

    actual = settings_service.get_environments_config(test_project)

    assert RASA_PRODUCTION_ENVIRONMENT not in actual["environments"]["rasa"]


def test_save_local_password(settings_service: SettingsService):
    password = "my password"

    assert not settings_service.get_community_user_password()

    settings_service.save_community_user_password(password)

    assert settings_service.get_community_user_password() == password


def test_update_local_password(settings_service: SettingsService):
    settings_service.save_community_user_password("password")

    updated_password = "updated password"
    settings_service.save_community_user_password(updated_password)

    assert settings_service.get_community_user_password() == updated_password


def test_inject_environments_config_from_file(settings_service, tmpdir):
    from rasa.utils.io import write_yaml_file

    environments_config = default_environments_config()["environments"]
    environments_file_path = os.path.join(tmpdir, "environments.yml")
    write_yaml_file(environments_config, environments_file_path)

    test_project = config.project_name
    settings_service.inject_environments_config_from_file(
        test_project, environments_file_path
    )

    # Get rid of the `ordereddict` types
    expected = json.loads(json.dumps(environments_config))

    assert (
        settings_service.get_environments_config(test_project)["environments"]
        == expected
    )


def test_failing_inject_environments_config_from_file(
    settings_service: SettingsService,
):
    test_project = config.project_name
    old_config = settings_service.get_environments_config(test_project)
    settings_service.inject_environments_config_from_file(
        test_project, "not existing path"
    )

    assert settings_service.get_environments_config(test_project) == old_config


def test_passing_inspect_config(settings_service: SettingsService):
    _config = """
    language: en
    pipeline: supervised_embeddings
    policies:
      - name: MemoizationPolicy
      - name: KerasPolicy
      - name: MappingPolicy
    """

    settings_service.inspect_and_load_yaml_config(_config)


def test_failing_inspect_config(settings_service: SettingsService):
    """Assert inspect_config() fails when config is invalid"""

    _config = """
    language: en
    policies:
      - name: MemoizationPolicy
      - name: KerasPolicy
      - name: MappingPolicy
    """

    with pytest.raises(InvalidConfigError):
        # noinspection PyProtectedMember
        settings_service.inspect_and_load_yaml_config(_config)


def test_inspect_and_save_yaml_config_from_request(
    settings_service: SettingsService, new_project: Text
):
    _config = b"""
    language: en
    pipeline: supervised_embeddings
    policies:
      - name: MemoizationPolicy
      - name: KerasPolicy
      - name: MappingPolicy
    """

    settings_service.inspect_and_save_yaml_config_from_request(
        _config, config.team_name, new_project
    )


@pytest.mark.parametrize(
    "model_config,error",
    [
        # simply invalid YAML
        (b"invalid: yaml: error", YAMLError),
        # missing 'policies' key
        (b"language: en\npipeline: supervised_embeddings", InvalidConfigError),
    ],
)
def test_inspect_and_save_yaml_config_from_request_yaml_error(
    model_config: bytes,
    error: Exception,
    settings_service: SettingsService,
    new_project: Text,
):
    with pytest.raises(error):
        settings_service.inspect_and_save_yaml_config_from_request(
            model_config, config.team_name, new_project
        )
