from typing import Dict, List, Optional

import pytest

import rasax.community.tracker_utils as tracker_utils
from rasa.core.actions.action import ACTION_LISTEN_NAME, ACTION_SESSION_START_NAME
from tests.unit.conftest import (
    action_event,
    bot_event,
    slot_set_event,
    TEST_USER_ADMIN,
    session_start_event,
)

session_start_sequence_with_slot = [
    action_event(TEST_USER_ADMIN, 0, action_name=ACTION_SESSION_START_NAME),
    session_start_event(TEST_USER_ADMIN, 1),
    slot_set_event(TEST_USER_ADMIN, "key", timestamp=2),
    bot_event(TEST_USER_ADMIN, 3),
    action_event(TEST_USER_ADMIN, 4, action_name=ACTION_LISTEN_NAME),
    bot_event(TEST_USER_ADMIN, 5),
    action_event(TEST_USER_ADMIN, 6, action_name=ACTION_LISTEN_NAME),
]

session_start_sequence_without_slot = [
    action_event(TEST_USER_ADMIN, 0, action_name=ACTION_SESSION_START_NAME),
    session_start_event(TEST_USER_ADMIN, 1),
    action_event(TEST_USER_ADMIN, 2, action_name=ACTION_LISTEN_NAME),
]

session_start_sequence_without_action_listen = [
    action_event(TEST_USER_ADMIN, 0, action_name=ACTION_SESSION_START_NAME),
    action_event(TEST_USER_ADMIN, 1, action_name=ACTION_LISTEN_NAME),
]


@pytest.mark.parametrize(
    "event,is_action_listen",
    [
        (action_event(TEST_USER_ADMIN, action_name=ACTION_LISTEN_NAME), True),
        (action_event(TEST_USER_ADMIN, action_name="another action"), False),
        (bot_event(TEST_USER_ADMIN), False),
        (slot_set_event(TEST_USER_ADMIN, key="key"), False),
    ],
)
def test_is_action_listen(event: Dict, is_action_listen: bool):
    assert tracker_utils.is_action_listen(event) == is_action_listen


@pytest.mark.parametrize(
    "event,is_action_session_start",
    [
        (action_event(TEST_USER_ADMIN, action_name=ACTION_SESSION_START_NAME), True),
        (action_event(TEST_USER_ADMIN, action_name="another action"), False),
        (bot_event(TEST_USER_ADMIN), False),
        (slot_set_event(TEST_USER_ADMIN, key="key"), False),
    ],
)
def test_is_action_session_start(event: Dict, is_action_session_start: bool):
    assert tracker_utils.is_action_session_start(event) == is_action_session_start


@pytest.mark.parametrize(
    "events,index",
    [
        ([], None),
        ([action_event(TEST_USER_ADMIN, action_name=ACTION_LISTEN_NAME)], None),
        (session_start_sequence_without_action_listen, None),
        (session_start_sequence_without_slot, 2),
        (session_start_sequence_with_slot, 4),
    ],
)
def test_index_of_session_start_sequence_end(events: List[Dict], index: Optional[int]):
    assert tracker_utils.index_of_session_start_sequence_end(events) == index


def test_remove_events_after_timestamp():
    timestamps = [
        0.04414652632108684,
        0.4955232397075445,
        0.49970819735303673,
        0.509080504858056,
        0.5264092149220423,
        0.5956426424708828,
        0.6611703339109208,
        0.6805125574093249,
        0.7244743522796836,
        0.9656514755973425,
    ]
    events = [
        action_event(
            TEST_USER_ADMIN, timestamp=timestamp, action_name=ACTION_SESSION_START_NAME
        )
        for timestamp in timestamps
    ]

    # remove events up to timestamp at index 2
    # we expect to get back events from index 3 onwards
    assert (
        tracker_utils.remove_events_until_timestamp(events, 0.49970819735303673)
        == events[3:]
    )


def test_remove_events_after_timestamp_no_events():
    timestamps = [0.04414652632108684, 0.4955232397075445, 0.49970819735303673]
    events = [
        action_event(
            TEST_USER_ADMIN, timestamp=timestamp, action_name=ACTION_SESSION_START_NAME
        )
        for timestamp in timestamps
    ]

    # removing events up to timestamp 1 means all events will be removed
    assert tracker_utils.remove_events_until_timestamp(events, 1) == []


@pytest.mark.parametrize(
    "events,timestamp",
    [
        ([], 0),
        ([bot_event(TEST_USER_ADMIN, 0)], 0),
        ([action_event(TEST_USER_ADMIN, 1, action_name=ACTION_LISTEN_NAME)], 1),
        (session_start_sequence_without_action_listen, 0),
        (session_start_sequence_without_slot, 2),
        (session_start_sequence_with_slot, 4),
    ],
)
def test_timestamp_of_conversation_start(
    events: List[Dict], timestamp: Optional[float]
):
    assert tracker_utils.timestamp_of_conversation_start(events) == timestamp


def test_remove_leading_action_session_start():
    tracker = {"events": session_start_sequence_with_slot}
    tracker_utils.remove_leading_action_session_start_from(tracker)

    assert tracker == {
        "events": [
            session_start_event(TEST_USER_ADMIN, 1),
            slot_set_event(TEST_USER_ADMIN, "key", timestamp=2),
            bot_event(TEST_USER_ADMIN, 3),
            action_event(TEST_USER_ADMIN, 4, action_name=ACTION_LISTEN_NAME),
            bot_event(TEST_USER_ADMIN, 5),
            action_event(TEST_USER_ADMIN, 6, action_name=ACTION_LISTEN_NAME),
        ]
    }
